/* Copyright (C) 2014 Dan Chapman <dpniel@ubuntu.com>

   This file is part of the dekko

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import QtQuick 2.0
import Ubuntu.Components 1.1
import Ubuntu.Components.ListItems 1.0 as ListItem
import Ubuntu.Components.Popups 1.0
import TrojitaCore 0.1
import DekkoCore 0.1
import "../Components"
import "../Dialogs"
import "../Utils/Utils.js" as Utils

BasePageWithComposer {
    id: oneMessagePage
    pageTitle: imapAccess.oneMessageModel.subject ? imapAccess.oneMessageModel.subject : messageHeader.fromAddress
    property url url: messageView.url
    property int uid
    property string mailboxName
    readonly property alias fromAddress: messageHeader.fromAddress
    property url externalLink
    property bool hasAttachments: false
    property bool displayingHTML: false

    //Composer properties
    property string fromName: i18n.tr("Reply to " + Utils.formatPlainArrayMailAddresses(imapAccess.oneMessageModel.from, true, false))
    composerReplyMode: true
    replySubject: imapAccess.oneMessageModel.subject
    replyMessage: ""// bind in PlainTextPart
    tipText: replyMode == ReplyMode.REPLY_NONE ? i18n.tr("Compose") : i18n.tr(fromName.length > 25 ? (fromName.slice(0,25) + "...") : fromName)
    // rfc states that a message should default to reply to the sender and require the MUA user
    // to change the reply mode. This is to reduce the inconvenience that not all people may want to
    // recieve the reply email.
    replyMode: ReplyMode.REPLY_PRIVATE



    Timer {
        id: markAsReadTimer
        repeat: false
        // TODO: make this configurable
        interval: 200
        onTriggered: imapAccess.oneMessageModel.markMessageAsRead()
    }

    Component.onCompleted: markAsReadTimer.start()

    Component {
        id: linkClickedDialog
        ConfirmationDialog {
            id: linkDialog
            title: i18n.tr("Open in browser?")
            text: i18n.tr("Confirm to open %1 in web browser".arg(externalLink))
            onConfirmClicked: Qt.openUrlExternally(externalLink)
        }
    }

    Action {
        id: backAct
        iconName: "back"
        onTriggered: {
            // We need to clear the model before going anywhere
            // otherwise the recipients in the model here will be present in the next composer
            submissionManager.recipientsModel.clear()
            pageStack.pop()
        }
    }
    Action {
        id: deleteAction
        iconName: "delete"
        visible: dekko.networkAvailable
        onTriggered: {
            imapAccess.markMessageDeleted(imapAccess.oneMessageModel.messageIndex, true)
            if (currentAccount.imapAutoExpungeOnDelete) {
                imapAccess.imapModel.expungeMailbox(imapAccess.msgListModel.currentMailbox())
            }
            backAct.trigger()
        }

    }
    pageHead: PageHeadConfiguration {
        id: oneMessagePageHead
        backAction: backAct
        actions: [deleteAction]
        sections {
            model: []
            selectedIndex: oneMessagePage.state === "panelHidden" ? oneMessagePage.head.sections.selectedIndex : -1
            onSelectedIndexChanged: {
                messageView.switchToSection(oneMessagePageHead.sections.selectedIndex);
            }
        }
    }

    flickable: null
    HeaderComponent {
        id: messageHeader
        anchors {
            left: parent.left
            right: parent.right
            top: parent.top
        }
    }
    Flickable {
        id: view
        anchors {
            top: messageHeader.bottom
            left: parent.left
            right: parent.right
            bottom: parent.bottom
        }

        interactive: !oneMessagePage.displayingHTML
        contentHeight: messageHeader.height + messageHeader.anchors.topMargin
                       + messageView.height + messageView.anchors.topMargin + units.gu(2)
        clip: true
        Column {
            id: messageView
            property var viewForEachSection: []
            property var sectionTypeForEachSection: []
            property real heightLeftForMessageContent: view.height - units.gu(4)
            property var backupSectionModel: [] // don't set head.sections.model unless there is more than two sections
            anchors {
                top: messageHeader.bottom
//                topMargin: units.gu(2)
                left: parent.left
                leftMargin: units.gu(2)
                rightMargin: units.gu(2)
                right: parent.right
            }
            Component.onCompleted: {
                partWidgetFactory.setNetAccessManager(imapAccess.msgQNAM);
                partWidgetFactory.setNetWatcher(imapAccess.networkWatcher);
                partWidgetFactory.create(imapAccess.oneMessageModel.messageFirstChildIndex,
                                         messageView,
                                         0,currentAccount.preferPlainText).parent = messageView;
            }
            function alertError(errorMessage) {
                console.log(errorMessage);
                PopupUtils.open(Qt.resolvedUrl("../Dialogs/InfoDialog.qml"), dekko, {
                                    title: i18n.tr("Error"),
                                    text: errorMessage
                                });
            }
            function openUrlExternally(url) {
                console.log("openUrlExternally " + url)
                if (url) {
                    oneMessagePage.externalLink = url
                    PopupUtils.open(linkClickedDialog)
                }
            }
            function addAttachmentItem(item) {
                console.log ("addAttachmentItem")
                if (!oneMessagePage.hasAttachments) {
                    messageView.addSection(attachmentView, SectionType.ATTACHMENT);
                    oneMessagePage.hasAttachments = true;
                }
                item.parent = attachmentView
            }
            function switchToSection(idx) {
                for (var i in viewForEachSection) {
                    if (i == idx) {
                        viewForEachSection[i].visible = true;
                        oneMessagePage.displayingHTML = (sectionTypeForEachSection[i] == SectionType.HTML);
                    } else {
                        viewForEachSection[i].visible = false;
                    }
                }
            }
            function addSection(item, type) {
                if (oneMessagePage.hasAttachments) {
                    //put before attachment section
                    var attachmentSectionIdex = sectionTypeForEachSection.indexOf(SectionType.ATTACHMENT);
                    backupSectionModel.splice(
                                attachmentSectionIdex, 0, Utils.sectionTypeToSectionName(type));
                    sectionTypeForEachSection.splice(
                                attachmentSectionIdex, 0, type);
                    messageView.viewForEachSection.splice(attachmentSectionIdex, 0, item);
                } else {
                    backupSectionModel.push(Utils.sectionTypeToSectionName(type));
                    sectionTypeForEachSection.push(type);
                    messageView.viewForEachSection.push(item);
                }
                if (backupSectionModel.length > 1) {
                    oneMessagePageHead.sections.model = backupSectionModel;
                    switchToSection(oneMessagePageHead.sections.selectedIndex)
                } else {
                    switchToSection(0);
                }

            }

            Column {
                id: attachmentView
                width: parent.width
                visible: false
            }
        }

        ComposerUtils {
            id: composerUtils
            // Force a reload here
            Component.onCompleted: oneMessagePage.reloadComposer()
        }

        Component {
            id: permissionDialog
            ConfirmationDialog {
                id: confDialog
                confirmAction: Action {
                    id: confAction
                    onTriggered: {
                        htmlPartView.permissionToDownload = true
                        htmlPartView.messageUrl = htmlPartView.messageUrl
                    }
                }
            }
        }
    }
}
