/* Copyright (C) 2006 - 2014 Jan Kundrát <jkt@flaska.net>
   Copyright (C) 2014 Dan Chapman <dpniel@ubuntu.com>

   This file is part of the Trojita Qt IMAP e-mail client,
   http://trojita.flaska.net/

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import QtQuick 2.0
import Ubuntu.Components 1.1
import Ubuntu.Components.ListItems 1.0 as ListItems
import TrojitaCore 0.1
import DekkoCore 0.1

Flickable {
    id: imapSettings
    anchors.fill: parent
    clip: true
    contentHeight: col.height + units.gu(5)

    signal imapSettingsChanged

    property Account account
    property alias imapUserName: imapUserNameInput.text
    property alias imapServer: imapServerInput.text
    property alias imapPort: imapPortInput.text
    property alias imapPassword: imapPasswordInput.text
    property alias imapSslModelIndex: connectionSelector.selectedIndex
    property bool settingsValid: usernameLabel.valid && hostLabel.valid
    property bool settingsModified: false

    function setupConnections() {
        settingsModified = false
        account.serverChanged.connect(settingModified)
        account.usernameChanged.connect(settingModified)
        account.portChanged.connect(settingModified)
        account.connMethodChanged.connect(settingModified)
        imapSettingsChanged.connect(dekko.settingsChanged)
    }

    function settingModified() { imapSettings.settingsModified = true }

    function saveImapSettings() {
        console.log("[LOG] Saving IMAP Settings..................")
        if (imapSettings.imapServer !== account.imapServer || imapSettings.imapUserName !== account.imapUsername) {
            console.log("[IMAP SETTINGS] Either the server or username has changed. Nuking old cache as it can no longer be used")
            imapAccess.nukeCache()
        }
        if (imapSettings.imapServer !== account.imapServer) {
            console.log("[IMAP SETTINGS] Ok so the server changed.... forgetting old ssl certificate if it exists")
            account.forgetSslCertificate()
            account.imapServer = imapSettings.imapServer
        }
        if (imapSettings.port !== account.imapPort) {
            account.imapPort = imapSettings.imapPort
        }
        if (imapSettings.imapUserName !== account.imapUsername) {
            account.imapUsername = imapSettings.imapUserName
        }
        if (connectionSelector.model.get(connectionSelector.selectedIndex).name !== account.imapSslMode) {
            account.imapSslMode = connectionSelector.model.get(connectionSelector.selectedIndex).name
        }
        console.log("[LOG] IMAP settings saved successfully")
    }

    states: [
        State {
            name: "NO_ACCOUNT"
            // Ensure the UI is clean
            PropertyChanges { target: imapSettings; imapServer: "" }
            PropertyChanges { target: imapServerInput; placeholderText: "" }
            PropertyChanges { target: imapSettings; imapUserName: "" }
            PropertyChanges { target: imapUserNameInput; placeholderText: "" }
            PropertyChanges { target: imapSettings; imapPassword: "" }
            PropertyChanges { target: imapPasswordInput; placeholderText: "" }
            PropertyChanges { target: imapSettings; imapSslModelIndex: 0 }
            PropertyChanges {
                target: imapSettings
                imapPort: connectionSelector.model.get(connectionSelector.selectedIndex).port
            }
        },
        State {
            name: "HAS_ACCOUNT"
            // Load user details
            PropertyChanges { target: imapSettings; imapServer: account.imapServer }
            PropertyChanges { target: imapSettings; imapUserName: account.imapUsername }
            PropertyChanges {
                target: imapSettings;
                imapPassword: imapPasswordWatcher.password ? imapPasswordWatcher.password : ""
            }
            PropertyChanges {
                target: imapSettings
                imapSslModelIndex: {
                    switch (account.imapSslMode) {
                    case "StartTLS":
                        1
                        break
                    case "SSL":
                        2
                        break
                    default:
                        0
                    }
                }
            }
            PropertyChanges {
                target: imapSettings
                imapPort: account.imapPort ?
                              account.imapPort :
                              connectionSelector.model.get(connectionSelector.selectedIndex).port
            }
        }
    ]

    Column {
        id: col
        spacing: units.gu(1)
        anchors {
            top: parent.top
            left: parent.left
            right: parent.right
            margins: units.gu(2)
        }

        OptionSelector {
            id: connectionSelector
            text: i18n.tr("Secure Connection")
            model: encryptionOptionsModel
            showDivider: false
            expanded: false
            colourImage: true
            delegate: connectionSelectorDelegate
            selectedIndex: imapSslModelIndex
            onSelectedIndexChanged: {
                imapPortInput.text = model.get(connectionSelector.selectedIndex).port
            }
        }

        Component {
            id: connectionSelectorDelegate
            OptionSelectorDelegate { text: description }
        }

        ListModel {
            id: encryptionOptionsModel
        }

        Label {
            id: hostLabel
            property bool valid: !!imapServerInput.text
            text: valid ? i18n.tr("Server address") : i18n.tr("Server address required")
            color: valid ? "#888888" : "red"
        }

        TextField {
            id: imapServerInput
            inputMethodHints: Qt.ImhNoAutoUppercase | Qt.ImhPreferLowercase | Qt.ImhUrlCharactersOnly | Qt.ImhNoPredictiveText
            anchors {
                left: col.left
                right: col.right
            }
            KeyNavigation.priority: KeyNavigation.BeforeItem
            KeyNavigation.tab: imapPortInput
            KeyNavigation.backtab: imapPasswordInput
        }

        Label {
            text: i18n.tr("Port")
        }
        Label {
            id: imapPortWarnLabel
            color: "red"
            visible: imapPortWarnLabel.text
        }
        TextField {
            id: imapPortInput
            inputMethodHints: Qt.ImhDigitsOnly
            validator: IntValidator { bottom: 1; top: 65535 }
            anchors {
                left: col.left
                right: col.right
            }
            onTextChanged: {
                if (parseInt(text) !== connectionSelector.model.get(connectionSelector.selectedIndex).port) {
                    imapPortWarnLabel.text = "This port is nonstandard. The default port is %1".arg(
                                connectionSelector.model.get(connectionSelector.selectedIndex).port)
                } else {
                    imapPortWarnLabel.text = ""
                }
            }
        }

        Label {
            id: usernameLabel
            property bool valid: !!imapUserNameInput.text
            text: valid ? i18n.tr("Username") : i18n.tr("Username required")
            color: valid ? "#888888" : "red"
        }
        TextField {
            id: imapUserNameInput
            anchors {
                left: col.left;
                right: col.right;
            }
            inputMethodHints: Qt.ImhNoAutoUppercase | Qt.ImhEmailCharactersOnly | Qt.ImhNoPredictiveText
            KeyNavigation.priority: KeyNavigation.BeforeItem
            KeyNavigation.tab: imapPasswordInput
        }

        Label {
            text: i18n.tr("Password")
        }
        TextField {
            id: imapPasswordInput
            inputMethodHints: Qt.ImhHiddenText | Qt.ImhNoAutoUppercase | Qt.ImhNoPredictiveText
            echoMode: TextInput.Password
            anchors {
                left: parent.left;
                right: parent.right;
            }
            KeyNavigation.priority: KeyNavigation.BeforeItem
            KeyNavigation.tab: imapServerInput
            KeyNavigation.backtab: imapUserNameInput
        }

        Label {
            id: passwordWarning
            anchors {
                left: parent.left
                right: parent.right
            }
            visible: imapPasswordInput.text.length > 0
            wrapMode: TextEdit.Wrap
            text: i18n.tr("This password will be saved in encrypted storage. " +
                          "If you do not enter password here, Dekko will prompt for one when needed.")
        }

    }


    Component.onCompleted: {
        encryptionOptionsModel.append({ "name": "No", "description": i18n.tr("No Encryption"), "port": 143 })
        encryptionOptionsModel.append({"name": "StartTLS", "description": i18n.tr("Use encryption (STARTTLS)"), "port": 143 })
        encryptionOptionsModel.append({ "name": "SSL", "description": i18n.tr("Force encryption (SSL/TLS)"), "port": 993 })
        account.imapSslMode ? state = "HAS_ACCOUNT" : state = "NO_ACCOUNT"
        setupConnections()
    }

    Connections {
        target: dekko
        onAutoConfigComplete: {
            console.log(config.imapHost + " " + config.imapPort + " " + config.imapMethod);
            imapUserName = username;
            imapPassword = password;
            imapServer = config.imapHost;
            for (var i = 0; i < encryptionOptionsModel.count; i++) {
                if (config.imapMethod === encryptionOptionsModel.get(i).name) {
                    imapSslModelIndex = i;
                    break;
                }
            }
            imapPort = config.imapPort;
        }
        onManualConfig: {
            imapUserName = username;
            imapPassword = password;
            imapServer = "";
            imapSslModelIndex = 0;
        }
    }
}
