/* Copyright (C) 2014 Dan Chapman <dpniel@ubuntu.com>

   This file is part of the Dekko

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Account.h"
#include <QCoreApplication>
#include <QSettings>
#include <QFile>
#include <QtQml/QQmlEngine>
#include <QUuid>
#include "Common/SettingsNames.h"
#include "Common/ConnectionMethod.h"
#include "Common/MetaTypes.h"
#include "Common/Paths.h"
#include "Common/PortNumbers.h"
#include "Plugins/PluginManager.h"
#include "AccountUtils.h"

namespace Dekko
{
namespace Accounts
{
Account::Account(QObject *parent) :
    QObject(parent), m_identityModel(0), m_accountName(QString()), m_imapPort(0), m_imapAutoExpunge(false),
    m_useSysProxy(false), m_imapEnableId(false), m_imapNeedsNetwork(true), m_cacheOfflineNumberOfDays(0),
    m_imapStartOffline(false), m_submissionPort(0), m_submissionAuthEnabled(false), m_submissionSaveToImap(false),
    m_submissionUseBurl(false), m_prefsAutoLoadImages(false), m_prefsPreferPlainText(true), m_devModeEnable(false),
    m_devModeImapLogging(false), m_devModeEnableThreading(false), m_devModeEnableContacts(false), m_useDefaultSignature(true)
{
    m_identityModel = new SenderIdentityModel(this);
    emit identityModelChanged();
}

QSettings *Account::settings()
{
    return m_settings;
}

void Account::setSettings(QSettings *settings)
{
    m_settings = settings;
}

QObject *Account::identityModel() const
{
    return m_identityModel;
}


QString Account::accountName() const
{
    return m_accountName;
}

void Account::setAccountName(const QString &name)
{
    if (name == m_accountName) {
        return;
    }
    m_accountName = name;
    emit accountNameChanged();
}

QString Account::description() const
{
    return m_description;
}

void Account::setDescription(const QString &description)
{
    if (description == m_description) {
        return;
    }
    m_description = description;
    emit descriptionChanged();
}

QString Account::accountImage() const
{
    return m_accountImagePath;
}

void Account::setAccountImage(const QString &image)
{
    if (image == m_accountImagePath) {
        return;
    }
    m_accountImagePath = image;
    emit accountImageChanged();
}

QString Account::imapServer() const
{
    return m_imapServer;
}

void Account::setImapServer(const QString &server)
{
    if (server == m_imapServer) {
        return;
    }
    m_imapServer = server;
    emit serverChanged();
}

int Account::imapPort() const
{
    return m_imapPort;
}

void Account::setImapPort(const int &port)
{
    if (port == m_imapPort) {
        return;
    }
    m_imapPort = port;
    emit portChanged();
}

QString Account::imapUsername() const
{
    return m_imapUserName;
}

void Account::setImapUsername(const QString &username)
{
    if (username == m_imapUserName) {
        return;
    }
    m_imapUserName = username;
    emit usernameChanged();
}

QString Account::imapSslMode() const
{
    switch (m_imapConnectionMethod) {
    case Common::ConnectionMethod::NetCleartext:
        return QLatin1String("No");
    case Common::ConnectionMethod::NetStartTls:
        return QLatin1String("StartTLS");
    case Common::ConnectionMethod::NetDedicatedTls:
        return QLatin1String("SSL");
    case Common::ConnectionMethod::Invalid:
    case Common::ConnectionMethod::Process:
        return QString();
    }

//    Q_ASSERT(false);
    return QString();
}

void Account::setImapSslMode(const QString &sslMode)
{
    if (sslMode == QLatin1String("No")) {
        setImapConnectionMethod(Common::ConnectionMethod::NetCleartext);
    } else if (sslMode == QLatin1String("SSL")) {
        setImapConnectionMethod(Common::ConnectionMethod::NetDedicatedTls);
    } else if (sslMode == QLatin1String("StartTLS")) {
        setImapConnectionMethod(Common::ConnectionMethod::NetStartTls);
    } else {
        Q_ASSERT(false);
    }
}

bool Account::imapAutoExpungeOnDelete() const
{
    return m_imapAutoExpunge;
}

void Account::setImapAutoExpungeOnDelete(const bool expunge)
{
    if (expunge == m_imapAutoExpunge) {
        return;
    }
    m_imapAutoExpunge = expunge;
    emit autoExpungeOnDeleteChanged();
}

QStringList Account::imapExtensionBlackList() const
{
    return m_imapExtBlacklist;
}

void Account::setImapExtensionBlackList(const QStringList &blacklist)
{
    if (blacklist == m_imapExtBlacklist) {
        return;
    }
    m_imapExtBlacklist = blacklist;
    m_imapExtBlacklist.removeDuplicates();
    emit extensionBlackListChanged();
}

QByteArray Account::imapPubKey() const
{
    return m_imapPubKey;
}

void Account::setImapPubKey(const QByteArray pubkey)
{
    if (pubkey == m_imapPubKey) {
        return;
    }
    m_imapPubKey = pubkey;
    emit pubKeyChanged();
}

bool Account::useSystemProxy() const
{
    return m_useSysProxy;
}

void Account::setUseSystemProxy(const bool useProxy)
{
    if (useProxy == m_useSysProxy) {
        return;
    }
    m_useSysProxy = useProxy;
    emit useSystemProxyChanged();
}

QStringList Account::imapProcessArgs() const
{
    return m_imapProcessArgs;
}

void Account::setImapProcessArgs(const QStringList &args)
{
    if (args == m_imapProcessArgs) {
        return;
    }
    m_imapProcessArgs = args;
    emit imapProcessArgsChanged();
}

bool Account::imapEnableId() const
{
    return m_imapEnableId;
}

void Account::setImapEnableId(const bool enable)
{
    if (enable == m_imapEnableId) {
        return;
    }
    m_imapEnableId = enable;
    emit imapEnableIdChanged();
}

bool Account::imapNeedsNetwork() const
{
    return m_imapNeedsNetwork;
}

void Account::setImapNeedsNetwork(const bool needsNetwork)
{
    if (needsNetwork == m_imapNeedsNetwork) {
        return;
    }
    m_imapNeedsNetwork = needsNetwork;
    emit imapNeedsNetworkChanged();
}

QString Account::cacheOffline() const
{
    return m_cacheOffline;
}

void Account::setCacheOffline(const QString &cacheOffline)
{
    if (cacheOffline == m_cacheOffline) {
        return;
    }
    m_cacheOffline = cacheOffline;
    emit cacheOfflineChanged();
}

int Account::cacheOfflineNumberOfDays() const
{
    return m_cacheOfflineNumberOfDays;
}

void Account::setCacheOfflineNumberOfDays(const int &numDays)
{
    if (numDays == m_cacheOfflineNumberOfDays) {
        return;
    }
    m_cacheOfflineNumberOfDays = numDays;
    emit cacheOfflineNumberOfDaysChanged();
}

bool Account::imapStartOffline() const
{
    return m_imapStartOffline;
}

void Account::setImapStartOffline(const bool startOffline)
{
    if (startOffline == m_imapStartOffline) {
        return;
    }
    m_imapStartOffline = startOffline;
    emit imapStartOfflineChanged();
}

void Account::forgetSslCertificate() const
{
    // FIXME
//    m_settings->remove(Common::SettingsNames::imapSslPemPubKey);

}

Common::ConnectionMethod Account::imapConnectionMethod() const
{
    return m_imapConnectionMethod;
}

void Account::setImapConnectionMethod(const Common::ConnectionMethod method)
{
    if (m_imapConnectionMethod == method) {
        return;
    }
    m_imapConnectionMethod = method;
    emit connMethodChanged();
}


QString Account::submissionServer() const
{
    return m_submissionServer;
}

void Account::setSubmissionServer(const QString &server)
{
    if (server == m_submissionServer) {
        return;
    }
    m_submissionServer = server;
    emit serverChanged();
}

int Account::submissionPort() const
{
    return m_submissionPort;
}

void Account::setSubmissionPort(const int &port)
{
    if (port == m_submissionPort) {
        return;
    }
    m_submissionPort = port;
    maybeShowPortWarning();
    emit portChanged();
}

QString Account::submissionUsername() const
{
    return m_submissionUsername;
}

void Account::setSubmissionUsername(const QString &username)
{
    if (username == m_submissionUsername) {
        return;
    }
    m_submissionUsername = username;
    emit submissionUsernameChanged();
}

bool Account::submissionAuthenticateEnabled() const
{
    return m_submissionAuthEnabled;
}

void Account::setSubmissionAuthenticateEnabled(const bool authEnabled)
{
    if (authEnabled == m_submissionAuthEnabled) {
        return;
    }
    m_submissionAuthEnabled = authEnabled;
    emit authenticateEnabledChanged();
}

QString Account::submissionPathToSendmail() const
{
    return m_submissionPathToSendmail;
}

void Account::setSubmissionPathToSendmail(const QString &path)
{
    if (path == m_submissionPathToSendmail) {
        return;
    }
    m_submissionPathToSendmail = path;
    emit pathToSendmailChanged();
}

bool Account::submissionSaveToImap() const
{
    return m_submissionSaveToImap;
}

void Account::setSubmissionSaveToImap(const bool save)
{
    if (save == m_submissionSaveToImap) {
        return;
    }
    m_submissionSaveToImap = save;
    emit saveToImapChanged();
}

QString Account::submissionSentMailboxName() const
{
    return m_submissionSentMailboxName;
}

void Account::setSubmissionSentMailboxName(const QString &mboxName)
{
    if (mboxName == m_submissionSentMailboxName) {
        return;
    }
    m_submissionSentMailboxName = mboxName;
    emit sentMailboxNameChanged();
}

bool Account::submissionUseBurl() const
{
    return m_submissionUseBurl;
}

void Account::setSubmissionUseBurl(const bool useBurl)
{
    if (useBurl == m_submissionUseBurl) {
        return;
    }
    m_submissionUseBurl = useBurl;
    emit useBurlChanged();
}

Composer::Submission::Method Account::submissionMethod() const
{
    return m_msaSubmissionMethod;
}

void Account::setSubmissionMethod(const Composer::Submission::Method &method)
{
    if (method == m_msaSubmissionMethod) {
        return;
    }
    switch (method) {
    case Composer::Submission::Method::SMTP:
    case Composer::Submission::Method::SSMTP:
    case Composer::Submission::Method::SMTP_STARTTLS:
        m_msaSubmissionMethod = method;
        setSubmissionPort(submissionDefaultPort(m_msaSubmissionMethod));
        break;
    case Composer::Submission::Method::SENDMAIL:
    case Composer::Submission::Method::IMAP_SENDMAIL:
        m_msaSubmissionMethod = method;
        break;
    }
    emit submissionMethodChanged();
}

int Account::submissionDefaultPort(const Composer::Submission::Method &method)
{
    switch (method) {
    case Composer::Submission::Method::SMTP:
    case Composer::Submission::Method::SMTP_STARTTLS:
        return Common::PORT_SMTP_SUBMISSION;
        break;
    case Composer::Submission::Method::SSMTP:
        return Common::PORT_SMTP_SSL;
        break;
    case Composer::Submission::Method::SENDMAIL:
    case Composer::Submission::Method::IMAP_SENDMAIL:
        break;
    }
    Q_ASSERT(false);
    return 0;
}

bool Account::autoLoadImages() const
{
    return m_prefsAutoLoadImages;
}

void Account::setAutoLoadImages(const bool autoLoad)
{
    if (autoLoad == m_prefsAutoLoadImages) {
        return;
    }
    m_prefsAutoLoadImages = autoLoad;
    emit autoLoadImagesChanged();
}

bool Account::preferPlainText() const
{
    return m_prefsPreferPlainText;
}

void Account::setPreferPlainText(const bool preferPlainText)
{
    if (preferPlainText == m_prefsPreferPlainText) {
        return;
    }
    m_prefsPreferPlainText = preferPlainText;
    emit preferPlainTextChanged();
}

bool Account::developerModeEnabled() const
{
    return m_devModeEnable;
}

void Account::setDeveloperModeEnabled(const bool enabled)
{
    if (enabled == m_devModeEnable) {
        return;
    }
    m_devModeEnable = enabled;
    emit developerModeEnabledChanged();
}

bool Account::devModeImapModelLogging() const
{
    return m_devModeImapLogging;
}

void Account::setDevModeImapModelLogging(const bool enabled)
{
    if (enabled == m_devModeImapLogging) {
        return;
    }
    m_devModeImapLogging = enabled;
    emit devModeImapModelLoggingChanged();
}

bool Account::devModeEnableThreading() const
{
    return m_devModeEnableThreading;
}

void Account::setDevModeEnableThreading(const bool enabled)
{
    if (enabled == m_devModeEnableThreading) {
        return;
    }
    m_devModeEnableThreading = enabled;
    emit devModeEnableThreadingChanged();
}

bool Account::devModeEnableContacts() const
{
    return m_devModeEnableContacts;
}

void Account::setDevModeEnableContacts(const bool enabled)
{
    if (enabled == m_devModeEnableContacts) {
        return;
    }
    m_devModeEnableContacts = enabled;
    emit devModeEnableContactsChanged();
}

bool Account::passwordMigrated() const
{
    return m_pwdMigrated;
}

void Account::setPasswordMigrated(const bool migrated)
{
    if (migrated == m_pwdMigrated) {
        return;
    }
    m_pwdMigrated = migrated;
    emit passwordMigrationChanged();
}

QString Account::generateNewAccountId() const
{
    m_settings->beginGroup("Accounts");
    // SO the idea here is we run a loop to get a unique id
    for (int i = 0; i < 10; i++) {
        // get a unique ID
        QUuid uid = QUuid::createUuid();
        // turn it to a string and strip {}
        QString uidString = uid.toString().remove(QLatin1Char('{')).remove(QLatin1Char('}'));
        if (!m_settings->childGroups().contains(uidString)) {
            return uidString;
        }
    }
    return QString();
}

void Account::reloadSettings()
{
    if (m_accountName.isEmpty()) {
        emit reloadFailed();
        return;
    }
    reloadSettings(m_settings, m_accountName);
}

void Account::reloadSettings(const QString &accountName)
{
    if (!m_settings) {
        emit reloadFailed();
        return;
    }
    reloadSettings(m_settings, accountName);
}

void Account::reloadSettings(QSettings *settings, const QString &accountName)
{

    qDebug() << "Reloading settings for account: " << accountName;
    m_settings = settings;
    m_accountName = accountName;
    m_settings->beginGroup(m_accountName);
    qDebug() << "Changed group to " << m_settings->group();
    m_description = m_settings->value("AccountDescription").toString();

    m_imapServer = m_settings->value(Common::SettingsNames::imapHostKey).toString();
    m_imapUserName = m_settings->value(Common::SettingsNames::imapUserKey).toString();
    if(!m_imapUserName.isEmpty()) {
        m_accountImagePath = AccountUtils::getDomainIcon(m_imapUserName);
    }
    if (m_settings->value(Common::SettingsNames::imapMethodKey).toString() == Common::SettingsNames::methodSSL) {
        m_imapConnectionMethod = Common::ConnectionMethod::NetDedicatedTls;
    } else if (m_settings->value(Common::SettingsNames::imapMethodKey).toString() == Common::SettingsNames::methodTCP) {
        m_imapConnectionMethod = m_settings->value(Common::SettingsNames::imapStartTlsKey).toBool() ?
                    Common::ConnectionMethod::NetStartTls : Common::ConnectionMethod::NetCleartext;
    } else if (m_settings->value(Common::SettingsNames::imapMethodKey).toString() == Common::SettingsNames::methodProcess) {
        m_imapConnectionMethod = Common::ConnectionMethod::Process;
    }
    m_imapPort = m_settings->value(Common::SettingsNames::imapPortKey, QVariant(0)).toInt();
    if (!m_imapPort) {
        switch (m_imapConnectionMethod) {
        case Common::ConnectionMethod::NetCleartext:
        case Common::ConnectionMethod::NetStartTls:
            m_imapPort = Common::PORT_IMAP;
            break;
        case Common::ConnectionMethod::NetDedicatedTls:
            m_imapPort = Common::PORT_IMAPS;
            break;
        case Common::ConnectionMethod::Process:
        case Common::ConnectionMethod::Invalid:
            // do nothing
            break;
        }
    }
    m_imapAutoExpunge = m_settings->value(Common::SettingsNames::imapAutoExpunge, false).toBool();
    m_imapExtBlacklist = m_settings->value(Common::SettingsNames::imapBlacklistedCapabilities).toStringList();
    m_imapPubKey = m_settings->value(Common::SettingsNames::imapSslPemPubKey).toByteArray();
    m_useSysProxy = m_settings->value(Common::SettingsNames::imapUseSystemProxy, true).toBool();
    m_imapProcessArgs = m_settings->value(Common::SettingsNames::imapProcessKey).toStringList();
    m_imapEnableId = m_settings->value(Common::SettingsNames::imapEnableId, true).toBool();
    m_imapNeedsNetwork = m_settings->value(Common::SettingsNames::imapNeedsNetwork, true).toBool();
    m_cacheOffline = m_settings->value(Common::SettingsNames::cacheOfflineKey).toString();
    m_cacheOfflineNumberOfDays = m_settings->value(Common::SettingsNames::cacheOfflineNumberDaysKey, QVariant(30)).toInt();
    m_imapStartOffline = m_settings->value(Common::SettingsNames::imapStartOffline, false).toBool();

    m_submissionServer = m_settings->value(Common::SettingsNames::smtpHostKey).toString();
    m_submissionUsername = m_settings->value(Common::SettingsNames::smtpUserKey).toString();
    QString connMethod = m_settings->value(Common::SettingsNames::msaMethodKey,
                                           Common::SettingsNames::methodSMTP).toString();
    if (connMethod == Common::SettingsNames::methodSMTP) {
        m_msaSubmissionMethod = m_settings->value(Common::SettingsNames::smtpStartTlsKey).toBool() ?
                            Composer::Submission::Method::SMTP_STARTTLS : Composer::Submission::Method::SMTP;
    } else if (connMethod == Common::SettingsNames::methodSSMTP) {
        m_msaSubmissionMethod = Composer::Submission::Method::SSMTP;
    } else if (connMethod == Common::SettingsNames::methodSSMTP) {
        m_msaSubmissionMethod = Composer::Submission::Method::SSMTP;
    } else if (connMethod == Common::SettingsNames::methodSENDMAIL) {
        m_msaSubmissionMethod = Composer::Submission::Method::SENDMAIL;
    } else if (connMethod == Common::SettingsNames::methodImapSendmail) {
        m_msaSubmissionMethod = Composer::Submission::Method::IMAP_SENDMAIL;
    } else {
        // unknown submission type, lets default to SMTP
        m_msaSubmissionMethod = Composer::Submission::Method::SMTP;
    }
    m_submissionPort = m_settings->value(Common::SettingsNames::smtpPortKey, QVariant(0)).toInt();
    if (!m_submissionPort) {
        switch (m_msaSubmissionMethod) {
        case Composer::Submission::Method::SMTP:
        case Composer::Submission::Method::SMTP_STARTTLS:
        case Composer::Submission::Method::SSMTP:
            // TODO add defaultPort()
            m_submissionPort = submissionDefaultPort(m_msaSubmissionMethod);
            break;
        case Composer::Submission::Method::SENDMAIL:
        case Composer::Submission::Method::IMAP_SENDMAIL:
            break;
        }
    }
    m_submissionAuthEnabled = m_settings->value(Common::SettingsNames::smtpAuthKey, false).toBool();
    m_submissionPathToSendmail = m_settings->value(Common::SettingsNames::sendmailKey,
                                         Common::SettingsNames::sendmailDefaultCmd).toString();
    m_submissionSaveToImap = m_settings->value(Common::SettingsNames::composerSaveToImapKey, false).toBool();
    m_submissionSentMailboxName = m_settings->value(Common::SettingsNames::composerImapSentKey, QLatin1String("Sent")).toString();
    m_submissionUseBurl = m_settings->value(Common::SettingsNames::smtpUseBurlKey, false).toBool();

    m_prefsPreferPlainText = m_settings->value(Common::SettingsNames::guiPreferPlaintextRendering, true).toBool();
    m_prefsAutoLoadImages = m_settings->value(Common::SettingsNames::imapAutoLoadImages, false).toBool();
    m_devModeEnable = m_settings->value(Common::SettingsNames::developerModeEnabled, false).toBool();
    m_devModeImapLogging = m_settings->value(Common::SettingsNames::enableImapModelLog, false).toBool();
    m_devModeEnableThreading = m_settings->value(Common::SettingsNames::devModeEnableThreading, false).toBool();
    m_devModeEnableContacts = m_settings->value(Common::SettingsNames::devModeEnableContacts, false).toBool();

    // TEMPORARY SENDER IDENTITY
    m_identityModel->clear();
    m_identityModel->reloadIdentities(m_settings);

    m_pwdMigrated = m_settings->value("PWD_MIGRATED", false).toBool();
    // and were done
    m_settings->endGroup();
    qDebug() << "RELOAD SUCCEEDED";
    emit reloadSucceeded();
}

QString Account::getDefaultSignature() const
{
    return QString(tr("Sent using Dekko from my Ubuntu device"));
}

void Account::reset()
{
    reloadSettings(m_settings, QString());
}

void Account::save()
{
    save(m_settings);
}

void Account::save(QSettings *settings)
{
    m_settings = settings;
    if (m_accountName.isEmpty()) {
        m_accountName = generateNewAccountId();
    }

    if (!m_settings->group().isEmpty()) {
        m_settings->endGroup();
    }

    if (m_settings->group() == "Accounts") {
        m_settings->endGroup();
    }

    if (m_settings->group() != m_accountName) {
        m_settings->beginGroup(m_accountName);
    }
    // NOTE: These are used by the AccountsListModel
    m_settings->setValue("AccountName", m_accountName);
    m_settings->setValue("AccountDescription", m_description);
//    m_settings->setValue("AccountImage", AccountUtils::getDomainIcon(m_imapUserName));

    m_settings->setValue(Common::SettingsNames::imapHostKey, m_imapServer);
    m_settings->setValue(Common::SettingsNames::imapUserKey, m_imapUserName);
    m_settings->setValue(Common::SettingsNames::imapPortKey, m_imapPort);
    m_settings->setValue(Common::SettingsNames::imapAutoExpunge, m_imapAutoExpunge);
    m_settings->setValue(Common::SettingsNames::imapBlacklistedCapabilities, m_imapExtBlacklist);
    switch (m_imapConnectionMethod) {
    case Common::ConnectionMethod::Invalid:
        break;
    case Common::ConnectionMethod::NetCleartext:
    case Common::ConnectionMethod::NetStartTls:
        m_settings->setValue(Common::SettingsNames::imapMethodKey, Common::SettingsNames::methodTCP);
        m_settings->setValue(Common::SettingsNames::imapStartTlsKey, m_imapConnectionMethod == Common::ConnectionMethod::NetStartTls);
        break;
    case Common::ConnectionMethod::NetDedicatedTls:
        m_settings->setValue(Common::SettingsNames::imapMethodKey, Common::SettingsNames::methodSSL);
        // Trying to communicate the fact that this is going to be an encrypted connection, even though
        // that settings bit is not actually used
        m_settings->setValue(Common::SettingsNames::imapStartTlsKey, true);
        break;
    case Common::ConnectionMethod::Process:
        m_settings->setValue(Common::SettingsNames::imapMethodKey, Common::SettingsNames::methodProcess);
        break;
    }
    m_settings->setValue(Common::SettingsNames::imapSslPemPubKey, m_imapPubKey);
    m_settings->setValue(Common::SettingsNames::imapUseSystemProxy, m_useSysProxy);
    m_settings->setValue(Common::SettingsNames::imapProcessKey, m_imapProcessArgs);
    m_settings->setValue(Common::SettingsNames::imapEnableId, m_imapEnableId);
    m_settings->setValue(Common::SettingsNames::imapNeedsNetwork, m_imapNeedsNetwork);
    m_settings->setValue(Common::SettingsNames::cacheOfflineKey, m_cacheOffline);
    m_settings->setValue(Common::SettingsNames::cacheOfflineNumberDaysKey, m_cacheOfflineNumberOfDays);
    m_settings->setValue(Common::SettingsNames::imapStartOffline, m_imapStartOffline);

    m_settings->setValue(Common::SettingsNames::smtpHostKey, m_submissionServer);
    switch (m_msaSubmissionMethod) {
    case Composer::Submission::Method::SMTP:
    case Composer::Submission::Method::SMTP_STARTTLS:
        m_settings->setValue(Common::SettingsNames::msaMethodKey, Common::SettingsNames::methodSMTP);
        m_settings->setValue(Common::SettingsNames::smtpStartTlsKey,
                             m_msaSubmissionMethod == Composer::Submission::Method::SMTP_STARTTLS); // unconditionally
        m_settings->setValue(Common::SettingsNames::smtpAuthKey, m_submissionAuthEnabled);
        m_settings->setValue(Common::SettingsNames::smtpUserKey, m_submissionUsername);
        m_settings->setValue(Common::SettingsNames::smtpHostKey, m_submissionServer);
        m_settings->setValue(Common::SettingsNames::smtpPortKey, m_submissionPort);
        break;
    case Composer::Submission::Method::SSMTP:
        m_settings->setValue(Common::SettingsNames::msaMethodKey, Common::SettingsNames::methodSSMTP);
        m_settings->setValue(Common::SettingsNames::smtpAuthKey, m_submissionAuthEnabled);
        m_settings->setValue(Common::SettingsNames::smtpUserKey, m_submissionUsername);

        m_settings->setValue(Common::SettingsNames::smtpPortKey, m_submissionPort);
        break;
    case Composer::Submission::Method::SENDMAIL:
        m_settings->setValue(Common::SettingsNames::msaMethodKey, Common::SettingsNames::methodSENDMAIL);
        m_settings->setValue(Common::SettingsNames::sendmailKey, m_submissionPathToSendmail);
        break;
    case Composer::Submission::Method::IMAP_SENDMAIL:
        m_settings->setValue(Common::SettingsNames::msaMethodKey, Common::SettingsNames::methodImapSendmail);
        break;
    }

    m_settings->setValue(Common::SettingsNames::composerSaveToImapKey, m_submissionSaveToImap);
    if (m_submissionSaveToImap) {
        m_settings->setValue(Common::SettingsNames::composerImapSentKey, m_submissionSentMailboxName);
        m_settings->setValue(Common::SettingsNames::smtpUseBurlKey, m_submissionUseBurl);
    } else {
        m_settings->setValue(Common::SettingsNames::smtpUseBurlKey, false);
    }

    m_settings->setValue(Common::SettingsNames::imapAutoLoadImages, m_prefsAutoLoadImages);
    m_settings->setValue(Common::SettingsNames::guiPreferPlaintextRendering, m_prefsPreferPlainText);
    m_settings->setValue(Common::SettingsNames::developerModeEnabled, m_devModeEnable);
    m_settings->setValue(Common::SettingsNames::enableImapModelLog, m_devModeImapLogging);
    m_settings->setValue(Common::SettingsNames::devModeEnableThreading, m_devModeEnableThreading);
    m_settings->setValue(Common::SettingsNames::devModeEnableContacts, m_devModeEnableContacts);

    // TEMPORARY SENDER IDENTITY
    m_identityModel->saveIdentites(m_settings);

    m_settings->setValue("PWD_MIGRATED", m_pwdMigrated);
    m_settings->endGroup();

    m_settings->beginGroup("Accounts");
    if (!m_settings->childGroups().contains(m_accountName)) {
        m_settings->beginGroup(m_accountName);
        m_settings->setValue("AccountName", m_accountName);
        m_settings->endGroup();
    }
    m_settings->endGroup();
    if (!QFile::exists(m_settings->fileName())) {
        m_settings->sync();
    }
    emit saved();
}

void Account::deleteAccountSettings()
{
    if (m_accountName.isEmpty())
        return;
    AccountUtils::nukeCache(m_accountName);
    m_settings->beginGroup("Accounts");
    m_settings->remove(m_accountName);
    m_settings->endGroup();
    m_settings->remove(m_accountName);
    emit accountSettingsDeleted();
}

void Account::nukeCurrentCache()
{
    AccountUtils::nukeCache(m_accountName);
}

void Account::maybeShowPortWarning()
{
    switch (m_msaSubmissionMethod) {
    case Composer::Submission::Method::SENDMAIL:
    case Composer::Submission::Method::IMAP_SENDMAIL:
        // this isn't a direct network connection -> ignore this
        return;
    case Composer::Submission::Method::SMTP:
    case Composer::Submission::Method::SMTP_STARTTLS:
    case Composer::Submission::Method::SSMTP:
        // we're connecting through the network -> let's check the port
        break;
    }
    QString portWarn;
    const int defPort = submissionDefaultPort(m_msaSubmissionMethod);
    if (m_submissionPort != defPort) {
        portWarn = QString("This port is nonstandard. The default port is %1").arg(defPort);
    }
    emit showPortWarning(portWarn);
}

void Account::appendExtensionToImapBlacklist(const QString &extension)
{
    if (m_imapExtBlacklist.contains(extension)) {
        return;
    }
    m_imapExtBlacklist.append(extension);
    m_imapExtBlacklist.removeDuplicates();
    emit extensionBlackListChanged();
}

void Account::removeExtensionFromBlackList(const QString &extension)
{
    if (!m_imapExtBlacklist.contains(extension)) {
        return;
    }
    m_imapExtBlacklist.removeOne(extension);
    emit extensionBlackListChanged();
}

}
}
