#include "UiUtils/PartWidgetFactory.h"
#include <QDebug>
#include <QQuickView>
#include <QQmlEngine>
#include <QQmlComponent>
#include <QQuickItem>
#include <QMimeType>
#include <QMimeDatabase>
#include "Imap/Model/ItemRoles.h"
#include "UiUtils/Formatting.h"
#include "app/Utils/Path.h"
#include "app/PartWidget/LoadablePartWidget.h"
#include "app/PartWidget/AttachmentView.h"
#include "app/PartWidget/SectionType.h"

extern QQuickView *qQuickViewer;

namespace UiUtils
{

QObject *PartWidgetFactory::newLabelWidget(QString message) {
    qDebug() << "newLabelWidget";
    QQmlEngine *engine = qQuickViewer->engine();
    QQmlComponent component(engine, Dekko::findQmlFile("qml/MessageView/LabelWidget.qml"));
    QQuickItem *item = qobject_cast<QQuickItem*>(component.create(engine->rootContext()));
    item->setProperty("text", QVariant(message));
    return item;
}

QObject *PartWidgetFactory::newLoadablePartWidget(QObject *parent, Imap::Network::MsgPartNetAccessManager *manager,
                                const QModelIndex &part, QObject *messageView,
                                UiUtils::PartWidgetFactory *factory, int recursionDepth,
                                const UiUtils::PartWidgetFactory::PartLoadingOptions loadingMode)
{
    qDebug() << "newLoadablePartWidget";
    return (new Dekko::PartWidget::LoadablePartWidget(
                qobject_cast<QQuickItem *>(messageView), part,
                factory, recursionDepth,
                loadingMode, qQuickViewer->engine()))
                    ->createItem();
}

void PartWidgetFactory::hideWidget(QObject *widget)
{
    qDebug() << "hideWidget";
    qobject_cast<QQuickItem *>(widget)->setVisible(false);
}

QObject *PartWidgetFactory::newAttachmentView(QObject *parent, Imap::Network::MsgPartNetAccessManager *manager,
                           const QModelIndex &partIndex, QObject *messageView,
                           QObject *contentWidget)
{
    qDebug() << "newAttachmentView";
    return (new Dekko::PartWidget::AttachmentView(
                qobject_cast<QQuickItem *>(messageView),
                manager, partIndex, qobject_cast<QQuickItem *>(contentWidget),
                qQuickViewer->engine()))->createItem();
}

QObject *PartWidgetFactory::newMultipartAlternativeWidget(QObject *parent, UiUtils::PartWidgetFactory *factory, const QModelIndex &partIndex,  QObject *messageView,
                                       const int recursionDepth, const UiUtils::PartWidgetFactory::PartLoadingOptions options)
{
    qDebug() << "newMultipartAlternativeWidget";
    QQmlEngine *engine = qQuickViewer->engine();
    QQmlComponent component(engine, Dekko::findQmlFile("qml/MessageView/MultipartAlternativeWidget.qml"));

    QQuickItem *columnItem = qobject_cast<QQuickItem*>(component.create(engine->contextForObject(messageView)));
    columnItem->setParent(messageView);

    const bool plaintextIsPreferred = options & UiUtils::PartWidgetFactory::PART_PREFER_PLAINTEXT_OVER_HTML;
    // Which "textual, boring part" should be shown?
    int preferredTextIndex = -1;

    // First loop iteration is used to find out what MIME type to show.
    // Two iterations are needed because we have to know about whether we're shown or hidden when creating child widgets.
    for (int i = 0; i < partIndex.model()->rowCount(partIndex); ++i) {
        QModelIndex anotherPart = partIndex.child(i, 0);
        Q_ASSERT(anotherPart.isValid());

        QString mimeType = anotherPart.data(Imap::Mailbox::RolePartMimeType).toString();

        const bool isPlainText = mimeType == QLatin1String("text/plain");
        const bool isHtml = mimeType == QLatin1String("text/html");

        // At first, check whether this is one of the textual parts which we like
        if (isPlainText && plaintextIsPreferred) {
            preferredTextIndex = i;
        } else if (isHtml && !plaintextIsPreferred) {
            preferredTextIndex = i;
        }
    }

    // Show that part which is "the most important". The choice is usually between text/plain and text/html, one of them will win,
    // depending on the user's preferences.  If there are additional parts, the user will be alerted about them later on.
    // As usual, the later parts win in general.
    int preferredIndex = preferredTextIndex == -1 ? partIndex.model()->rowCount(partIndex) - 1 : preferredTextIndex;

    // The second loop actually creates the widgets
    for (int i = 0; i < partIndex.model()->rowCount(partIndex); ++i) {
        QModelIndex anotherPart = partIndex.child(i, 0);
        Q_ASSERT(anotherPart.isValid());
        // TODO: This is actually not perfect, the preferred part of a multipart/alternative
        // which is nested as a non-preferred part of another multipart/alternative actually gets loaded here.
        // I can live with that.
        QObject *item = factory->create(anotherPart, messageView, recursionDepth + 1,
                                        filteredForEmbedding(i == preferredIndex ?
                                            options :
                                            options | UiUtils::PartWidgetFactory::PART_IS_HIDDEN));
        QString mimeType = anotherPart.data(Imap::Mailbox::RolePartMimeType).toString();

        const bool isPlainText = mimeType == QLatin1String("text/plain");
        const bool isHtml = mimeType == QLatin1String("text/html");

        if (isPlainText) {
            //: Translators: use something very short, perhaps even "text". Don't describe this as "Clear text".
            //: This string is used as a caption of a tab showing the plaintext part of a mail which is e.g.
            //: sent in both plaintext and HTML formats.
            mimeType = tr("Plaintext");
        } else if (isHtml) {
            //: Translators: caption of the tab which shows a HTML version of the mail. Use some short, compact text here.
            mimeType = tr("HTML");
        }
        QMetaObject::invokeMethod(columnItem, "addMimeType", Q_ARG(QVariant, QVariant::fromValue(item)), Q_ARG(QVariant, mimeType));

    }

    QMetaObject::invokeMethod(columnItem, "setDisplayedMessageType", Q_ARG(QVariant, QVariant(preferredIndex)));

    return columnItem;
}

QObject *PartWidgetFactory::newMultipartSignedWidget(QObject *parent, UiUtils::PartWidgetFactory *factory, const QModelIndex &partIndex,
                                       const int recursionDepth, const UiUtils::PartWidgetFactory::PartLoadingOptions options)
{
    qDebug() << "TODO newMultipartSignedWidget";
    return newLabelWidget("Viewing multipart/signed message is not supported yet");
}

QObject *PartWidgetFactory::newGenericMultipartWidget(QObject *parent, UiUtils::PartWidgetFactory *factory, const QModelIndex &partIndex, QObject *messageView,
                                   const int recursionDepth, const UiUtils::PartWidgetFactory::PartLoadingOptions options)
{
    qDebug() << "newGenericMultipartWidget";
    // multipart/mixed or anything else, as mandated by RFC 2046, Section 5.1.3
    QQmlEngine *engine = qQuickViewer->engine();
    QQmlComponent component(engine, Dekko::findQmlFile("qml/MessageView/GenericMultipartWidget.qml"));

    QQuickItem *columnItem = qobject_cast<QQuickItem*>(component.create(engine->contextForObject(messageView)));
    columnItem->setParent(messageView);

    for (int i = 0; i < partIndex.model()->rowCount(partIndex); ++i) {
        using namespace Imap::Mailbox;
        QModelIndex anotherPart = partIndex.child(i, 0);
        Q_ASSERT(anotherPart.isValid()); // guaranteed by the MVC
        QObject *res = factory->create(anotherPart, messageView, recursionDepth + 1, filteredForEmbedding(options));
        if (res != NULL) {
            if (res->property("isAttachmentView").toBool()) {
                QMetaObject::invokeMethod(messageView, "addAttachmentItem", Q_ARG(QVariant, QVariant::fromValue(qobject_cast<QQuickItem *>(res))));
            } else {
                qobject_cast<QQuickItem *>(res)->setParentItem(columnItem);
            }
        }
    }
    return columnItem;
}

QObject *PartWidgetFactory::newMessage822Widget(QObject *parent, UiUtils::PartWidgetFactory *factory, const QModelIndex &partIndex,
                             const int recursionDepth, const UiUtils::PartWidgetFactory::PartLoadingOptions options)
{
    qDebug() << "TODO newMessage822Widget";
    return newLabelWidget("Viewing message/rfc822 message is not supported yet");
}

QObject *PartWidgetFactory::newSimplePartWidget(QObject *parent, Imap::Network::MsgPartNetAccessManager *manager,
                             const QModelIndex &partIndex, QObject *messageView, bool dontRegisterAsSection)
{
    qDebug() << "newSimplePartWidget";
    QQmlEngine *engine = qQuickViewer->engine();
    bool isPlainText = false;
    bool isHtml = false;
    QQmlComponent *component;
    if (partIndex.data(Imap::Mailbox::RolePartMimeType).toString() == QLatin1String("text/html")){
        component = new QQmlComponent(engine, Dekko::findQmlFile("qml/MessageView/HtmlPart.qml"));
        isHtml = true;
    } else if (partIndex.data(Imap::Mailbox::RolePartMimeType).toString().startsWith(QLatin1String("image"))) {
        component = new QQmlComponent(engine, Dekko::findQmlFile("qml/MessageView/ImagePart.qml"));
    } else {
        component = new QQmlComponent(engine, Dekko::findQmlFile("qml/MessageView/PlainTextPart.qml"));
        isPlainText = true;
    }
    QUrl url;
    url.setScheme(QLatin1String("trojita-imap"));
    url.setHost(QLatin1String("msg"));
    url.setPath(partIndex.data(Imap::Mailbox::RolePartPathToPart).toString());

    QQuickItem *item = qobject_cast<QQuickItem*>(component->create(engine->contextForObject(messageView)));

    item->setParent(messageView);
    component->deleteLater();

    if (partIndex.data(Imap::Mailbox::RolePartMimeType).toString() == QLatin1String("text/html"))
        connect(item, SIGNAL(linkClicked(QVariant)), messageView, SLOT(openUrlExternally(QVariant)));

    if (isPlainText || isHtml) {
        item->setProperty("msgPartIndex", partIndex);
    } else {
        item->setProperty("source", url);
    }

    if ((isHtml || isPlainText) && !dontRegisterAsSection) {
        if (isHtml) {
            QMetaObject::invokeMethod(messageView, "addSection",
                Q_ARG(QVariant, QVariant::fromValue(item)),
                Q_ARG(QVariant, QVariant::fromValue((int)Dekko::PartWidget::SectionType::HTML)));
        } else {
            //plainText
            QMetaObject::invokeMethod(messageView, "addSection",
                Q_ARG(QVariant, QVariant::fromValue(item)),
                Q_ARG(QVariant, QVariant::fromValue((int)Dekko::PartWidget::SectionType::PLAIN_TEXT)));
        }
    }

    return item;
}

}
